@@ -4,21 +4,21 @@ import inspect
 import io
 import json
 import logging
-import os
-import uuid
 
 import jsonschema
+import structlog
 
-from .connection import ConnectionCache
-from .dispatcher import Stdout, Http
-from .exceptions import (
+from insightconnect_plugin_runtime.connection import ConnectionCache
+from insightconnect_plugin_runtime.dispatcher import Stdout, Http
+from insightconnect_plugin_runtime.exceptions import (
     ClientException,
     ServerException,
     LoggedException,
     ConnectionTestException,
     PluginException,
 )
-from .metrics import MetricsBuilder
+from insightconnect_plugin_runtime.metrics import MetricsBuilder
+from insightconnect_plugin_runtime.util import is_running_in_cloud
 
 message_output_type = {
     "action_start": "action_event",
@@ -30,21 +30,21 @@ message_output_type = {
 
 class Workflow(object):
     def __init__(
-            self,
-            shortOrgId=None,
-            orgProductToken=None,
-            uiHostUrl=None,
-            jobId=None,
-            stepId=None,
-            versionId=None,
-            nextStepId=None,
-            nextEdgeId=None,
-            triggerId=None,
-            jobExecutionContextId=None,
-            time=None,
-            connectionTestId=None,
-            connectionTestTimeout=None,
-            workflowId=None,
+        self,
+        shortOrgId=None,
+        orgProductToken=None,
+        uiHostUrl=None,
+        jobId=None,
+        stepId=None,
+        versionId=None,
+        nextStepId=None,
+        nextEdgeId=None,
+        triggerId=None,
+        jobExecutionContextId=None,
+        time=None,
+        connectionTestId=None,
+        connectionTestTimeout=None,
+        workflowId=None,
     ):
         """
         Worflow object for the Meta Class
@@ -108,7 +108,7 @@ class Workflow(object):
 
 
 class Meta(object):
-    """ Meta properties for a plugin """
+    """Meta properties for a plugin"""
 
     def __init__(self, name="", vendor="", description="", version="", workflow=None):
         self.name, self.vendor, self.description, self.version, self.workflow = (
@@ -135,14 +135,14 @@ class Plugin(object):
     """A Komand Plugin."""
 
     def __init__(
-            self,
-            name="",
-            vendor="",
-            description="",
-            version="",
-            connection=None,
-            custom_encoder=None,
-            custom_decoder=None,
+        self,
+        name="",
+        vendor="",
+        description="",
+        version="",
+        connection=None,
+        custom_encoder=None,
+        custom_decoder=None,
     ):
         self.name = name
         self.vendor = vendor
@@ -163,28 +163,28 @@ class Plugin(object):
         self.custom_encoder = custom_encoder
 
     def add_trigger(self, trigger):
-        """ add a new trigger """
+        """add a new trigger"""
         self.triggers[trigger.name] = trigger
 
     def add_action(self, action):
-        """ add a new action """
+        """add a new action"""
         self.actions[action.name] = action
 
     def add_task(self, task):
-        """ add a new task"""
+        """add a new task"""
         self.tasks[task.name] = task
 
     def envelope(
-            self,
-            message_type,
-            input_message,
-            log,
-            success,
-            output,
-            error_message,
-            state,
-            ex: None,
-            is_test: False
+        self,
+        message_type,
+        input_message,
+        log,
+        success,
+        output,
+        error_message,
+        state,
+        ex: None,
+        is_test: False,
     ):
         """
         Creates an output message of a step's execution.
@@ -227,25 +227,27 @@ class Plugin(object):
                     output_message["exception"] = {
                         "cause": "Plugin connection test failed.",
                         "assistance": "See error log for more details.",
-                        "data": ex.__repr__()
+                        "data": ex.__repr__(),
                     }
 
                 # Build the metrics blob to attach to the output payload
                 # Only supported in cloud environments right now
-                if os.environ.get("PLUGIN_RUNTIME_ENVIRONMENT") == "cloud":
-                    metrics_builder = MetricsBuilder(plugin_name=self.name,
-                                                     plugin_version=self.version,
-                                                     plugin_vendor=self.vendor,
-                                                     input_message=input_message,
-                                                     exception_=ex,
-                                                     workflow_id=self.connection.meta.workflow.workflowId,
-                                                     org_id=self.connection.meta.workflow.shortOrgId)
+                if is_running_in_cloud():
+                    metrics_builder = MetricsBuilder(
+                        plugin_name=self.name,
+                        plugin_version=self.version,
+                        plugin_vendor=self.vendor,
+                        input_message=input_message,
+                        exception_=ex,
+                        workflow_id=self.connection.meta.workflow.workflowId,
+                        org_id=self.connection.meta.workflow.shortOrgId,
+                    )
                     output_message["metrics"] = metrics_builder.build()
 
         return {"body": output_message, "version": "v1", "type": message_type}
 
     def marshal(self, msg, fd):
-        """ Marshal a message to fd. """
+        """Marshal a message to fd."""
 
         if self.custom_encoder is None:
             json.dump(msg, fd)
@@ -254,7 +256,7 @@ class Plugin(object):
         fd.flush()
 
     def unmarshal(self, fd):
-        """ Unmarshal a message. """
+        """Unmarshal a message."""
 
         if self.custom_decoder is None:
             msg = json.load(fd)
@@ -347,12 +349,14 @@ class Plugin(object):
         if input_message_meta is None:
             input_message_meta = {}
         self.connection.meta.set_workflow(input_message_meta)
-        request_id = uuid.uuid4()
+
+        # Add StreamHandler to record plugin action/trigger/task logs for output back to consumer
+        # log_stream will record logs via StreamHandler and be included in the plugin output
         log_stream = io.StringIO()
         stream_handler = logging.StreamHandler(log_stream)
-        logger = logging.getLogger("step_handler_{}".format(request_id))
-        logger.setLevel("DEBUG" if is_debug else "INFO")
-        logger.addHandler(stream_handler)
+        stream_handler.setLevel(logging.DEBUG if is_debug else logging.INFO)
+        logging.getLogger("plugin").addHandler(stream_handler)
+        logger = structlog.get_logger("plugin")
 
         success = True
         caught_exception = None
@@ -364,7 +368,12 @@ class Plugin(object):
             # Attempt to grab message type first
             message_type = input_message.get("type")
             out_type = message_output_type.get(message_type)
-            if message_type not in ["action_start", "trigger_start", "task_start", "connection_test"]:
+            if message_type not in [
+                "action_start",
+                "trigger_start",
+                "task_start",
+                "connection_test",
+            ]:
                 raise ClientException(
                     'Unsupported message type "{}"'.format(message_type)
                 )
@@ -422,7 +431,7 @@ class Plugin(object):
                     log_stream,
                     is_test,
                     is_debug,
-                    is_connection_test=True
+                    is_connection_test=True,
                 )
         except (ClientException, ServerException, PluginException, Exception) as e:
             success = False
@@ -438,15 +447,21 @@ class Plugin(object):
                 str(caught_exception),
                 state,
                 caught_exception,
-                is_test
+                is_test,
             )
-            print(f"Output is: {output}")
             if not success:
                 raise LoggedException(caught_exception, output)
             return output
 
     def start_step(
-            self, message_body, step_key, logger, log_stream, is_test=False, is_debug=False, is_connection_test=False
+        self,
+        message_body,
+        step_key,
+        logger,
+        log_stream,
+        is_test=False,
+        is_debug=False,
+        is_connection_test=False,
     ):
         """
         Starts an action.
@@ -475,7 +490,7 @@ class Plugin(object):
                     "returning any content (no connection test function)"
                 )
             output = func()
-            
+
             return output
 
         else:
@@ -508,7 +523,7 @@ class Plugin(object):
 
             params = message_body["input"]
             state = {}
-    
+
             if not is_test:
                 # Validate input message
                 try:
@@ -516,15 +531,19 @@ class Plugin(object):
                     if step_key == "task":
                         state = message_body["state"]
                         step.state.validate(state)
-    
+
                     # Validate required inputs
                     # Step inputs will be checked against schema for required properties existence
                     # This is needed to prevent null/empty string values from being passed as output to input of steps
                     step.input.validate_required(params)
                 except jsonschema.exceptions.ValidationError as e:
-                    raise ClientException("{} input JSON was invalid".format(step_key), e)
+                    raise ClientException(
+                        "{} input JSON was invalid".format(step_key), e
+                    )
                 except Exception as e:
-                    raise Exception("Unable to validate {} input JSON".format(step_key), e)
+                    raise Exception(
+                        "Unable to validate {} input JSON".format(step_key), e
+                    )
 
             # Log step information for improved debugging with users
             step.logger.info(
@@ -535,7 +554,7 @@ class Plugin(object):
                     step_name=step.name,
                 )
             )
-        
+
             if is_test:
                 # Check if connection test func available. If so - use it (preferred). Else fallback to action/trigger test
                 if hasattr(step.connection, "test"):
@@ -544,7 +563,7 @@ class Plugin(object):
                     func = step.test
             else:
                 func = step.run
-    
+
             # Backward compatibility with steps with missing argument params
             # The SDK has always defined the signature of the run/test methods to include the params dictionary.
             # However, the code generation generates the test method without the params argument.
@@ -559,13 +578,13 @@ class Plugin(object):
                     output, state = func()
                 else:
                     output = func()
-    
+
             # Don't validate output for any test functions - action/trigger tests shouldn't be validated due to them
             # not providing value and a connection test shouldn't be validated due to it being generic/universal
             if not is_test:
                 step.output.validate(output)
-    
+
             if step_key == "task" and not is_test:
                 return output, state
-        
+
             return output
